@@ -4,6 +4,7 @@ import os
 import re
 import sys
 import time
+import datetime
 import traceback
 import html2text
 from bs4 import BeautifulSoup
@@ -45,8 +46,8 @@ class GPTWebElement(webdriver.remote.webelement.WebElement):
 class GPTSeleniumAgent:
     def __init__(
         self,
-        instructions,
-        chromedriver_path,
+        instructions="",
+        chromedriver_path=None,
         chrome_options={},
         user_data_dir="user_data",
         headless=False,
@@ -89,7 +90,10 @@ class GPTSeleniumAgent:
             instruction_output_file is None
             or instruction_output_file.endswith(".yaml")
             or instruction_output_file.endswith(".json")
-        ), "Instruction output file must be a YAML or JSON file or None."
+        ), "Instruction output file must be a YAML or JSON file or None." 
+        assert (
+            chromedriver_path is not None
+        ), "Please provide a path to the chromedriver executable."
         self.model_for_instructions = model_for_instructions
         self.model_for_responses = model_for_responses
         self.instruction_output_file = instruction_output_file
@@ -114,6 +118,12 @@ class GPTSeleniumAgent:
         """Set up the driver."""
         _chrome_options = webdriver.ChromeOptions()
         _chrome_options.add_argument(f"user-data-dir={user_data_dir}")
+        # 🤫 Evade detection. 
+        # https://stackoverflow.com/questions/53039551/selenium-webdriver-modifying-navigator-webdriver-flag-to-prevent-selenium-detec
+        _chrome_options.add_argument('--disable-blink-features=AutomationControlled')
+        _chrome_options.add_experimental_option("excludeSwitches", ["enable-automation"])
+        _chrome_options.add_experimental_option('useAutomationExtension', False)
+
         self.headless = headless
         if headless:
             _chrome_options.add_argument("--headless")
@@ -123,6 +133,9 @@ class GPTSeleniumAgent:
         # Instantiate Service with the path to the chromedriver and the options.
         service = Service(chromedriver_path)
         self.driver = webdriver.Chrome(service=service, options=_chrome_options)
+        # 🤫 Evade detection.
+        self.driver.execute_script("Object.defineProperty(navigator, 'webdriver', {get: () => undefined})")
+
 
     """Helper functions"""
 
@@ -210,6 +223,16 @@ class GPTSeleniumAgent:
         elements = [ele for ele in elements if ele.attrs]
         return elements
 
+    def __complete(self):
+        """What to run when the agent is done."""
+        if self.enable_memory:
+            current_time = int(time.time())
+            fpath = "memory_{current_time}.json".format(current_time=current_time)
+            self.memory.save(fpath)
+
+        if self.close_after_completion:
+            self.driver.quit()
+
     def __run_compiled_instructions(self, instructions):
         """Runs Python code previously compiled by InstructionCompiler."""
         ldict = {"env": self}
@@ -219,8 +242,7 @@ class GPTSeleniumAgent:
         except:
             self.__handle_agent_exception(instructions)
 
-        if self.close_after_completion:
-            self.driver.quit()
+        self.__complete()
 
     def __print_instruction_and_action(self, instruction, action):
         """Logging the instruction and action."""
@@ -238,9 +260,10 @@ class GPTSeleniumAgent:
         class_name = self.__class__.__name__
         stack_trace = stack_trace.replace(class_name, "env")
         # Get the number after the word "line " in the stack trace.
-        line_num = int(
-            stack_trace.split("line ")[1].split(",")[0].strip()
-        )
+        everything_after_line = stack_trace.split("line ")[1]
+        everything_before_rest = everything_after_line.split(",")[0]
+        line_num_str = everything_before_rest.split()[0].strip()
+        line_num = int(line_num_str)
         return {"stack_trace": stack_trace, "line_num": line_num}
 
     def __save_html_snapshot(self):
@@ -339,8 +362,7 @@ class GPTSeleniumAgent:
                 self.instruction_output_file
             )
 
-        if self.close_after_completion:
-            self.driver.quit()
+        self.__complete()
 
     def __switch_to_element_iframe(func):
         """Decorator function to switch to the iframe of the element."""
@@ -389,7 +411,7 @@ class GPTSeleniumAgent:
         if not url.startswith("http"):
             url = "http://" + url
         self.driver.get(url)
-        time.sleep(2)
+        time.sleep(1)
         if self.enable_memory:
             # Get all the visible text from the page and add it to the memory.
             text = self.get_text_from_page(entire_page=False)
@@ -437,7 +459,14 @@ class GPTSeleniumAgent:
         self.driver.switch_to.default_content()
 
     def find_element(self, by="id", value=None):
-        return self.find_elements(by, value)[0]
+        found_elements = self.find_elements(by, value)
+        if len(found_elements) == 0:
+            raise Exception("No elements found.")
+
+        # Iterate through until the first displayed one is found, then return that one.
+        for element in found_elements:
+            if element.is_displayed():
+                return element
 
     def find_elements(self, by="id", value=None):
         """Wrapper over `driver.find_elements` which also scans iframes.
@@ -452,12 +481,19 @@ class GPTSeleniumAgent:
         and in all iframes. Returns a list of GPTWebElement objects.
         """
         elements = self.driver.find_elements(by, value)
+        # Screen for elements that are not displayed.
+        elements = [element for element in elements if element.is_displayed()]
         elements = [GPTWebElement(element) for element in elements]
+        # Find all iframes on the page and switch to each one to find
+        # their elements.
         iframes = self.driver.find_elements(by=By.TAG_NAME, value="iframe")
-        logger.info("Found {num} iframes.".format(num=len(iframes)))
+        logger.debug("Found {num} iframes.".format(num=len(iframes)))
         for iframe in iframes:
             self.driver.switch_to.frame(iframe)
             iframe_elements = self.driver.find_elements(by, value)
+            iframe_elements = [
+                element for element in iframe_elements if element.is_displayed()
+            ]
             iframe_elements = [
                 GPTWebElement(element, iframe=iframe) for element in iframe_elements
             ]
